/*
*	
*	ls -l | hfmt '%s %s %s %8s %s %3s %6s <a href="%U">%H</a>\n' 1 4 5 6 7 8 9 10 10
*
*/

#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>

/*
 * print dates in the format
 * Wkd, DD Mon YYYY HH:MM:SS GMT
 * parse dates of formats
 * Wkd, DD Mon YYYY HH:MM:SS GMT
 * Weekday, DD-Mon-YY HH:MM:SS GMT
 * Wkd Mon ( D|DD) HH:MM:SS YYYY
 * plus anything similar
 */
static char *
weekdayname[7] =
{
	"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"
};
static char *
wdayname[7] =
{
	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

static char *
monname[12] =
{
	"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

int xflag=0;


int
hdatefmt(Fmt *f)
{
	Tm *tm;
	ulong t;

	t = va_arg(f->args, ulong);
	tm = gmtime(t);
	return fmtprint(f, "%s, %.2d %s %.4d %.2d:%.2d:%.2d GMT",
		wdayname[tm->wday], tm->mday, monname[tm->mon], tm->year+1900,
		tm->hour, tm->min, tm->sec);
}


int
hurlfmt(Fmt *f)
{
	char *b;
	char *s, *t, *e;
	int n, status;
	char *unsafe=" ;?#:@!&=+$,<>%\"{}|\\^[]()*`'";
	/* 	unsafe chars in path that should be escaped
	*	see the comment in "httpd/libhttpd/urlfmt.c" in Pegasus 2.3
	*	Note that ";?#" is included here	*/

	s = va_arg(f->args, char*);

	n = 0;
	t = s;
	while(*t){
		if((xflag && !isascii(*t))||iscntrl(*t)||strchr(unsafe,*t))
			n += 3;
		else
			n++;
		t++;
	}
	n++;
	b = malloc(n);
	t = b;
	e = t + n;
	while(*s){
		if((xflag && !isascii(*s))|| iscntrl(*s)||strchr(unsafe,*s))
			t = seprint(t,e,"%%%02X", (uchar)*s);
		else
			*t++ = *s;
		s++;
	}
	*t = 0;
	status = fmtstrcpy(f, b);
	free(b);
	return status;
}


int
htmlfmt(Fmt *f)
{
	char *hspace=" \t\n";
	char *b;
	char *s, *t, *e;
	int n, status;
	s = va_arg(f->args, char*);

	/* we count first the number of '&', '<' and '>' */
	n = 0;
	t = s;
	while(*t){
		if(*t == '&' || *t == '<' || *t == '>' || strchr(hspace,*t))
			n += 6;
		else
			n++;
		t++;
	}
	n++;

	b = malloc(n);
	t = b;
	e = t + n;
	while(*s){
		if(*s == '&')
			t = seprint(t,e,"&amp;");
		else if(*s == '<')
			t = seprint(t,e,"&lt;");
		else if(*s == '>')
			t = seprint(t,e,"&gt;");
		else if(strchr(hspace,*s))
			t = seprint(t,e,"&#%03d;",*s);
		else
			*t++ = *s;
		s++;
	}
	*t = 0;
	status = fmtstrcpy(f, b);
	free(b);
	return status;
}


void
usage(void)
{
	fprint(2,"usage: hfmt [-a] format args ....\n");
	exits("usage");
}

#define size 32

enum{
	STR = 0,
	INT,
	FLO,
};

void
Bvtprint(Biobufhdr *bp, char *fmt, char **args)
{
	int d, i;
	char c, *s, *t;
	/* determine the data type of n-th field */
	i = 0;
	s = t = fmt;
	while(*s){
		if(*s != '%'){
			s++;
			continue;
		}
		*s = 0;
		Bprint(bp, "%s", t);
		*s = '%';
		t = s;
		s++;
		if(*s == 0)
			break;
		if(*s == '%'){
			Bprint(bp,"%%");
			s++;
			t = s;
			continue;
		}

		while(*s && (*s == '.' || isdigit(*s)))
			s++;
		if(*s == 0)
			break;
		d = 0;
		if(strchr("cdiDoxXu",*s))
			d = INT;
		else if(strchr("eEgGf\n",*s))
			d = FLO;
		else if(strchr("sUH", *s))
			d = STR;
		s++;
		c = *s;
		if(args[i] == nil)
			break;
		*s = 0;
		switch(d){
		case INT: Bprint(bp, t, atoi(args[i])); break;
		case STR: Bprint(bp, t, args[i]); break;
		case FLO: Bprint(bp, t, atof(args[i])); break;
		}
		i++;
		*s = c;
		t = s;
	}
	Bprint(bp, "%s", t);
}


void
main(int argc, char *argv[])
{
	Biobuf ib, ob;
	char *args[size], *v[size];
	int ind[size];
	int m=size;
	char *fmt,*line, *s, *t;
	int i,n, max;
	int aflag=0;

	ARGBEGIN{
	case 'a': aflag++; break;
	case 'x': xflag++; break;
	default: usage();
	}ARGEND

	if(argc == 0)
		usage();
	fmt = argv[0];
	s = t = fmt;
	while(*s){
		if(*s == '\\'){
			s++;
			switch(*s){
			case 't': *t = '\t'; break;
			case 'n': *t = '\n'; break;
			case 'r': *t = '\r'; break;
			case '\\': *t = '\\'; break;
			case '?': *t = '\?'; break;
			case '"': *t = '\"'; break;
			default: *t = *s;
			}
		}
		else
			*t = *s;
		s++;
		t++;
	}
	*t = 0;
	//print("DEBUG: %s", fmt);

	argc--;
	argv++;

	fmtinstall('D', hdatefmt);
	fmtinstall('U', hurlfmt);
	fmtinstall('H', htmlfmt);

	if(aflag){
		for(i = 0; i < argc; i++){
			v[i] = argv[i];
		}
		for(i = argc; i < size; i++)
			v[i] = nil;

//		print(fmt,v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],v[10],
//			v[11],v[12],v[13],v[14],v[15]);
		Binit(&ob, 1, OWRITE);
		Bvtprint(&ob, fmt,v);
		exits(nil);
	}

	max = 0;
	for(i = 0; i < argc; i++){
		ind[i] = n = atoi(argv[i]);
		if(n > max)
			max = n;
	}

	/*
	print("DEBUG: ");
	for(i = 0; i < argc; i++)
		print("%d",ind[i]);
	print("\n");
	*/

	Binit(&ib, 0, OREAD);
	Binit(&ob, 1, OWRITE);

	while(line = Brdstr(&ib,'\n', 0)){/* assign = */
		n = tokenize(line, args, m);
		//for(i = 0; i < n; i++)
		//	print("DEBUG: %s\n",args[i]);
		if(n < max)
			sysfatal("fields too short");
		for(i = 0; i < argc; i++)
			v[i] = args[ind[i] - 1];
		for(i = argc; i < size; i++)
			v[i] = nil;
		/* 	I don't know how to handle variable print
		*	It seems va_list no help */
//		print(fmt,v[0],v[1],v[2],v[3],v[4],v[5],v[6],v[7],v[8],v[9],v[10],
//			v[11],v[12],v[13],v[14],v[15]);
		Bvtprint(&ob, fmt,v);
	}

	exits(nil);
}


/*	Experiment
term% 8.hfmt -a 'abc %% %s=%d %s=%4.2f %D\n' alice 18 bob 23.789 `{date -n}
abc % alice=18 bob=23.79 Fri, 25 Aug 2006 08:58:07 GMT
term% 
term% 8.hfmt -a '%d %s=%6.5g\n' -18 pi 3.14
-18 pi=  3.14
term% 
term% 8.hfmt -a '%s %s %s\n' alice
alice %s %s
term% 
term% 8.hfmt -a '%s\n' alice bob carol
alice
term% 
*/
