#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ndb.h>
#include <ctype.h>
#include "misc.h"

extern int lockouttime;
static void nsop(int argc, char *argv[]);

void*
ezalloc(ulong n)
{
	void *p;

	p = malloc(n);
	if(p == nil)
		sysfatal("out of memory");
	memset(p, 0, n);
	return p;
}

char*
estrdup(char *s)
{
	s = strdup(s);
	if(s == nil)
		sysfatal("out of memory");
	return s;
}


int
xstrcmp(char *s, char *t)
{
	if(s == nil){
		if(t)
			return -1;
		return 0;
	}
	if(t)
		return strcmp(s,t);
	return 1;
}

char*
trim(char *s)
{
	char *t;
	while(isspace(*s))
		s++;
	t = strchr(s, 0);
	t--;
	while(isspace(*t ))
		t--;
	t++;
	*t = 0;
	return s;
}

char *
strlower(char *s)
{
	char *s0;
	s0 = s;
	while(*s){
		*s = tolower(*s);
		s++;
	}
	return s0;
}

/*
*	test if the path is deeper then or equal to ref
*	deeper(path, ref) returns matched size
*		0 if path is not deeper
*		matched size if deeper # the size is used in basic_auth()
*	example:
*	# path	ref
*	/abc		/abc			# return 4
*	/abc		/abc/		# return 4
*	/abc/	/abc			# return 4
*	/abc/	/abc/		# return 4
*	/abc		/abcdef		# return 0
*	/abc		/abc/def		# return 0
*	/abc/	/abcdef		# return 0
*	/abc/def	/abc/		# return 4
*	/abc/def	/abc			# return 4
*	/abc		/			# return 1
*	a		a			# return 1
*	a/		a			# return 1
*	a/		a/			# return 1
*
*	path is:
*	-	absolute path
*	-	relative path
*	-	canonical uri of the form: scheme://host:port/path
*/
int
deeper(char *path, char *ref)
{
	int n;
	int m = 0;
	int f = 0;
	n = strlen(ref);
	if(n == 1 && ref[0] == '/' && path[0] == '/')
		return 1;
	if(ref[n-1] == '/'){
		ref[n-1] = 0;
		n--;
		f = 1;
	}
	if(strncmp(path, ref, n) == 0 && (path[n] == 0 || path[n] == '/'))
		m = n;
	if(f)
		ref[n] = '/';
	return m;
}


#ifdef NOTE
*	shell style path matching function
*	I modified to simplify original one
*	and made some modification of the matching rule.
*	now, '/' is a simplly regular charactor except
*	"/*/" is matched to "/".
*	ref: /sys/src/cmd/rc/glob.c
*	-Kenar-
*
 * 	Does the string s match the pattern p
 * 	* matches any sequence of characters
 * 	? matches any single character
 * 	[...] matches the enclosed list of characters
 *	~ in the beginning of [...] means compliment
 *	- in [...] means range in ascii order
 *
 *	return value
 *	1: matched
 *	0: not matched
 #endif
int
match(char *s, char *p)
{
	char *s0;
	int compl, hit, lo, hi, t, c;
	s0 = s;
	for(;*p; s++,p++){
		switch(*p){
		case '*':
			++p;
			if(*p == '/' && s > s0)
				if(match(s-1, p)) return 1;
			for(;;){
				if(match(s, p)) return 1;
				if(!*s) break;
				s++;
			}
			return 0;
		case '?':
			if(*s=='\0') return 0;
			break;
		case '[':
			if(*s=='\0') return 0;
			c=*s;
			p++;
			compl=*p=='~';
			if(compl) p++;
			hit=0;
			while(*p!=']'){
				if(*p=='\0') return 0;		/* syntax error */
				lo=*p;
				p++;
				if(*p!='-') hi=lo;
				else{
					p++;
					if(*p=='\0') return 0;	/* syntax error */
					hi=*p;
					p++;
					if(hi<lo){ t=lo; lo=hi; hi=t; }
				}
				if(lo<=c && c<=hi) hit=1;
			}
			if(compl) hit=!hit;
			if(!hit) return 0;
			break;
		default:
			if(*p != *s) return 0;
		}
	}
	return *s=='\0';
}


char *
skip(char *s, char *p)
{
	while(*s && strchr(p, *s))
		s++;
	return s;
}

char *
skipto(char *s, char *p)
{
	while(*s && !strchr(p, *s))
		s++;
	return s;
}

char *
skipapha(char *s)
{
	while(*s && isalpha(*s))
		s++;
	return s;
}


/*	getword get a word from s and returns point
*	next to search ( the point is next char of the word)
*	if there is no word, it returns nil
*	-Kenar-	*/
char *
getword(char *s, char *w, int size)
{
	char *r;
	r = w + size;
	s = skip(s, " \t");
	if(*s == 0)
		return nil;
	while(*s && !strchr(" \t", *s)){
		if(w == r){
			w--;
			break;
		}
		*w++ = *s++;
	}
	*w = 0;
	return s;
}

/*
*	Pegasus before 2.2
*	executable returns 1
*	if the file is executable for `others'
*	else returns 0
*
*	Pegasus 2.2
*	executable returns 1
*	if the file is executable for the httpd
*	else returns 0
*/
int
executable(char *path)
{
/*	// Pegasus before 2.2
	Dir *dir;
	dir = dirstat(path);
	if(dir == nil)
		return 0;
	if(dir->mode & DMEXEC)
		return 1;
	free(dir);
	return 0;
*/
	return !access(path, AEXEC);
}

int
waitfor(int pid, char *msg)
{	int cpid;
	Waitmsg *w;
	
	while((w = wait()) != nil){
		if(pid == w->pid){
			strncpy(msg, w->msg, ERRMAX);
			free(w);
			return 0;
		}
		free(w);
	}
	return -1;
}

/*	bind only addns
*	original source: /sys/src/libauth/newns.c	*/
#define NARG	5
int
chgns(char *file)
{
	Biobuf *spec;
	int argc;
	char *cmd, *argv[NARG];

	spec = Bopen(file, OREAD);
	if(spec == 0){
		werrstr("can't open %s: %r", file);
		return -1;
	}
	while(cmd = Brdline(spec, '\n')){
		cmd[Blinelen(spec)-1] = '\0';
		while(*cmd==' ' || *cmd=='\t')
			cmd++;
		if(*cmd == '#')
			continue;
		/* we need not support '#' tree	*/
		argc = getfields(cmd, argv, NARG, 1, " \t");
		if(argc)
			nsop(argc, argv);
	}
	Bterm(spec);
	return 0;
}

/* 	bind only nsop
*	original source: /sys/src/libauth/newns.c	*/
static void
nsop(int argc, char *argv[])
{
	char *argv0;
	ulong flags;

	flags = 0;
	argv0 = 0;
	ARGBEGIN{
	case 'a':
		flags |= MAFTER;
		break;
	case 'b':
		flags |= MBEFORE;
		break;
	case 'c':
		flags |= MCREATE;
		break;
	case 'C':
		// flags |= MCACHE;
		break;
	}ARGEND

	if(!(flags & (MAFTER|MBEFORE)))
		flags |= MREPL;
	if(strcmp(argv0, "bind") == 0 && argc == 2 && *argv[0] != '#'){
		/*	Adding some security protections
		*	system files should not be exposed to /doc	*/
		argv[1] = cleanname(argv[1]);
		if(strcmp(argv[1], "/doc") == 0 || strncmp(argv[1], "/doc/",4) == 0){
			argv[0] = cleanname(argv[0]);
			if(strcmp(argv[0],"/doc") != 0 && strncmp(argv[0], "/doc/", 4) != 0)
				return;
		}
		bind(argv[0], argv[1], flags);
	}
}

static char *
cat3(char *buf, char *s1, char *s2, char *s3)
{
	strcpy(buf, s1); strcat(buf, s2); strcat(buf, s3);
	return buf;
}

/*
*	nconnect
*	usage:
*		int tot;
*		n = nconnect("202.250.160.40", &tot);
*	return:
*		n: no. of connection from the ip
*		tot: total connection
*
*/
int
nconnect(char *ip, int *tot)
{
	int fd, i, k, n, m, ndir;
	Dir *db, *dir;
	char buf[64];
	char *p;
	fd = open("/net/tcp", OREAD);
	if(fd == -1)
		return -1;
	ndir = dirreadall(fd,&db);
	close(fd);
	if(ndir == -1){
		free(db);
		return -1;
	}
	m = 0; // total connection by the ip
	k = 0;   // total open connection
	for(i=0; i<ndir; i++){
		fd = open(cat3(buf, "/net/tcp/", db[i].name,"/status"), OREAD);
		if(fd < 0)
			continue;
		n = read(fd, buf, sizeof buf);
		close(fd);
		if(n <= 0)
			continue;
		buf[n-1] = 0;
		p = strchr(buf,' ');
		if(p)
			*p = 0;
		if(strcmp(buf,"Closed") == 0)
			continue;
		fd = open(cat3(buf, "/net/tcp/", db[i].name,"/remote"), OREAD);
		if(fd < 0)
			continue;
		n = read(fd, buf, sizeof buf);
		close(fd);
		if( n <= 0)
			continue;
		if(buf[0] == ':')
			continue;
		buf[n-1] = 0;
		p = strchr(buf, '!');
		if(p == nil)
			continue;
		*p = 0;
		if(strcmp(buf, ip) == 0)
			m++;
		k++;
	}
	free(db);
	if(tot)
		*tot = k;
	return m;
}

/*	stolen from /sys/src/cmd/aux/listen1.c	*/
char*
remoteaddr(char *dir)
{
	static char buf[128];
	char *p;
	int n, fd;

	snprint(buf, sizeof buf, "%s/remote", dir);
	fd = open(buf, OREAD);
	if(fd < 0)
		return "";
	n = read(fd, buf, sizeof(buf));
	close(fd);
	if(n > 0){
		buf[n] = 0;
		p = strchr(buf, '!');
		if(p)
			*p = 0;
		return buf;
	}
	return "";
}

/*
*	blacklist
*/
#define NBLIST	10
typedef struct BlackIP BlackIP;
struct BlackIP
{
	char *ip;
	long time; // registered time
};

static BlackIP blacklist[NBLIST];
static int nblist;
/*	examine if the ip is in blacklist	*/
int
black(char *ip)
{
	int n, k;
	long now;
	now = time(nil);
	n = nblist - NBLIST;
	if(n < 0)
		n = 0;
	for(k = nblist -1; k >= n; k--)
		if(strcmp(blacklist[k % NBLIST].ip, ip) == 0
		&& blacklist[k % NBLIST].time + lockouttime > now)
			return 1;
	return 0;
}

void
putblist(char *ip)
{
	int n;
	n = nblist % NBLIST;
	if(blacklist[n].ip){
		free(blacklist[n].ip);
		blacklist[n].ip = nil;
	}
	blacklist[n].ip = estrdup(ip);
	blacklist[n].time = time(nil);
	nblist++;
}


/*
*	history
*/
#define NHLIST	100
typedef struct hist hist;
struct hist {
	char *ip;
	int time; // time in unit of 0.1 sec since first access
};

static hist histlist[NHLIST];
static int nhlist;
static vlong ht0;
void
puthlist(char *ip)
{
	int n;
	if(ht0 == 0)
		ht0 = nsec();
	n = nhlist % NHLIST;
	if(histlist[n].ip)
		free(histlist[n].ip);
	histlist[n].ip = estrdup(ip);
	histlist[n].time = (nsec() - ht0)/100000000; // 0.1 sec
	nhlist++;
}

/*	a simple hash function
*	we will make error in 1/256 or more
*	sorry that remote ip if we made error
*	we are in emergency	*/
static int
puthash(char *ip)
{
	int sum, i;
	static int count[256];
	if(ip == nil){
		for(i = 0; i < 256; i++)
			count[i] = 0;
		return 0;
	}
	sum = 0;
	while(*ip){
		if(*ip != '.'){
			sum << 1;
			sum ^= 07 & (*ip - '0');
		}
		ip++;
	}
	sum &= 0xff; // hash
	count[sum] += 1;
	return count[sum];
}

#define inithash() puthash(nil)

/*	detect the ip from the ips in last t sec	*/
char *
arrest(int t)
{
	int n, k, i, j, m;
	char *ip;
	int t0;
	if(nhlist == 0)
		return nil;
	n = nhlist - NHLIST;
	if(n < 0)
		n = 0;
	t0 = (nsec() - ht0) /100000000; // change the unit to 0.1 sec
	t = 10*t; // change the unit to 0.1 sec
	m = 0;
	ip = nil;
	inithash();
	for(k = nhlist -1; k >= n; k--){
		i = k % NHLIST;
		if(t0 - histlist[i].time < t){
			j = puthash(histlist[i].ip);
			//syslog(0,"http", "arrest: i=%d ip=%s j=%d", i, histlist[i].ip, j);
			if(j > m){
				m = j;
				ip = histlist[i].ip;
			}
		}
	}
	return ip;
}


/*
*	httpd configuration
*/

static Db *configlist;

int
readconfig(char *file)
{
	char *line, *s;
	Biobuf *b;
	char *field[3];

	b = Bopen(file, OREAD);
	if(b == nil)
		return 0; // need not exist
	while((line = Brdline(b, '\n')) != nil){
		line[Blinelen(b)-1] = 0;
		s = strchr(line, '#');
		if(s != nil)
			*s = '\0'; 	/* chop comment */
		if(getfields(line, field, 3, 1, " \t") < 2)
			continue;
		if(db_put(&configlist, field[0], field[1]) < 0)
			return -1;
	}
	Bterm(b);
	return 0;
}

void
setintconfig(int *var, char *name)
{
	char *p;
	p = db_get(configlist,name);
	if(p)
		*var = atoi(p);
}

void
setstrconfig(char **var, char *name)
{
	char *p;
	p = db_get(configlist,name);
	if(p)
		*var = p;
}


char *
getdom(char *attr, char *val)
{
	Ndb *db;
	Ndbtuple *t, *nt;
	Ndbs s;
	static char buf[128];
	char *v;
	v = nil;
	db = ndbopen(nil);
	t = ndbsearch(db, &s, attr, val);
	while(t){
		for(nt = t; nt; nt = nt->entry)
			if(strcmp(nt->attr, "dom") == 0){
				v = strncpy(buf, nt->val, sizeof buf);
				break;
			}
		ndbfree(t);
		if(v)
			break;
		t = ndbsnext(&s, attr, val);
	}
	ndbclose(db);
	return v;
}

